/*
    v0.01 - original
    v0.02 - loads from patch_loader.ini at startup
          - scroll bar on tfp list


 - place .tfp files in /ProgramFiles
 - supports up to 32 .tfp files
 - patch_loader.ini
   e.g.
    debug=0     [0=off, 1=debug]
    patch_0=RecRoundel.tfp
    patch_0=DreamFree.tfp
    ...
    patch_31=TAPdiskFix.tfp

*/
//
//requires libfirebird
//
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <tap.h>
#include "exTap.h"

//#define type_hook_registers void
//#define TYPE_exTapEvent void
//#define is_firmware(p) ((((dword)p&0xfc000000)==0x80000000) ? 1 : 0)

#include "tf5800.h"
#include "libFirebird.h"

//if no extap
word fw_sysid;
dword fw_gp;
dword fw_tap_table;

#ifdef _WIN32
#else
#define ID_USERTAP 0xbdb02c02
TAP_ID( ID_USERTAP );
TAP_PROGRAM_NAME("patch_loader");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("v0.02 patch_loader");
TAP_ETCINFO(__DATE__);
#endif

#define bitrev16(d)  (((d << 8) & 0xff00) + ((d >> 8) & 0x00ff))
#define bitrev32(d)  (((d << 24) & 0xff000000) + ((d << 8) & 0xff0000) + ((d >> 8) & 0xff00) + ((d >> 24) & 0x00ff))

#define LE16(d)       bitrev16(d)
#define LE32(d)       bitrev32(d)
#define BE16(d)      (d)
#define BE32(d)      (d)

int debug = 0;
int remove_patch = 0;

//----------------------------------------------------------------------------
//
//firmware dependant
//


//----------------------------------------------------------------------------
#define MAX_SEGMENTS 0x10

typedef struct {
    dword   Address;                    //use 0 to scan the whole binary
    dword   CodeLength;                 //in dword instructions
    dword   OriginalCode[0x200];
    dword   OriginalMask[0x200];        //not used if Address<>0
    dword   PatchedCode[0x200];
    dword   PatchMask[0x200];
}type_PatchSegment;

typedef struct {
    dword   DestinationGP;              //unaligned      littleendian           //use 0 for all firmware levels
    dword   NrOfSegments;               //unaligned      littleendian
    char    Comment[0x50];
    type_PatchSegment   PatchSegment[MAX_SEGMENTS];
}type_PatchDetails;

typedef struct {
    dword   Magic;
    dword   FileVersion;
    char    Author[0x50];
    char    Name[0x50];
    char    email[0x50];
    char    EnglishDescr[0x100];
    char    GermanDescr[0x100];
    char    Version;
    dword   NrOfPatches;                //unaligned      bigendian  ???
    type_PatchDetails   PatchDetails;
}type_PatchModule;


type_PatchModule PatchModule;
char tmenu[0x20][0x64];


int read_tfp(char *filename, type_PatchModule *PatchModule, dword gp, int remove, int dump);
int find_patch(type_PatchSegment *PatchSegment);
int check_patch(type_PatchSegment *PatchSegment);

//----------------------------------------------------------------------------
void oprintf(const void *fmt, ...);
void ffprintf(TYPE_File *Op, const void *fmt, ...);

//hooks
int hook1_enabled = 0;
int hook1_id = -2;
int apply_hook1();
int remove_hook1();
dword hook1(dword param1, dword param2);

typedef struct {
    dword caller_addr;
    dword callee_addr;
    dword count;
}type_profile;
type_profile *profile = 0;

#define MAX_ENTRIES 0x10000
int num_entries = 0;

void *profile_hash = 0;

void patch_profile_hook();
int capture_end = 0;

//----------------------------------------------------------------------------
//menu stuff
char *tap_name;
type_menu main_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {11, "xxx",    },

        {0, "Load Patch", },
        {1, "Remove Patch", },
        {2, "", },
        {3, "", },
        {4, "", },
        {5, "", },
        {6, "", },
        {7, "", },
        {8, "", },
        {9, "", },

        {10, "Exit Tap", },
    },
};

type_menu tfp_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {1, "xxx",    },

        {0, "load tfp file", },
    },
};
//----------------------------------------------------------------------------
//
//tsr_commander stuff
//
#ifdef _WIN32
#else
#include "TSRCommander.inc"

bool TSRCommanderExitTAP()
{
    close();
    return FALSE;
}

void TSRCommanderConfigDialog()
{
    ShowMenu(&main_menu);
}

//extap
//int is_tsr_commander_running(void)
//{
//    int i;
//    type_TAPProcess *TAPProcess;
//
//    for (i=0; i<16; i++) {
//        TAPProcess = (type_TAPProcess *)(fw_tap_table + (i*32));
//        if (TAPProcess->firmwaregp) {
//            if (TAPProcess->header) {
//                if (TAPProcess->header->id == 0x807a10b2) {  //tsr_commander
//                    return 1;
//                }
//                if (TAPProcess->header->id == 0x81001021) {  //tap_commander
//                    return 1;
//                }
//            }
//        }
//    }
//    return 0;
//}

//firebirdlib
int is_tsr_commander_running(void)
{
    int i;
    tTAPTable *TAPProcess;

    for (i=0; i<16; i++) {
        TAPProcess = (tTAPTable *)(fw_tap_table + (i*32));
        if (TAPProcess->FWgp) {
            if (TAPProcess->TAPHeader) {
                if (TAPProcess->TAPHeader->TAPID == 0x807a10b2) {  //tsr_commander
                    return 1;
                }
                if (TAPProcess->TAPHeader->TAPID == 0x81001021) {  //tap_commander
                    return 1;
                }
            }
        }
    }
    return 0;
}
#endif

//----------------------------------------------------------------------------
int TAP_Main(void)
{
    int ret;

    tap_name =  (char *)__tap_program_name__;
    sprintf(strLog, "Opening %s", tap_name);
    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);

//    //determine support for different firmwares
//    ret = exTAP_Init();
//    TAP_Print("\nexTAP_Init(), ret=%x", ret);
//    if (ret) {
//        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
//        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
//        return 0;
//    }

    //initialise firebirdlib
    ret = InitTAPex();
//    TAP_Print("\nInitTAPex(), ret=%x", ret);
    if (ret == 0) {
        sprintf(strLog, "%s(), InitTAPex(), ret=%x", __func__, ret);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
        return 0;
    }
    fw_sysid = SysID;
    fw_gp = FWgp;
    fw_tap_table = TAP_TableAddress;

    ini_read();

    main_menu.visible = 0;
    tfp_menu.visible = 0;

    TSRCommanderInit(0, TRUE);
    return 1;                 //0=application mode, 1=tsr mode
}

int close(void)
{
    DeleteMenu(&main_menu);
    DeleteMenu(&tfp_menu);

    sprintf(strLog, "Closing  %s", tap_name);
    log_message(LOG_SERIAL|LOG_SCREEN_WAIT, strLog, DELAY_SHORT);

    if (message_rgn) TAP_Osd_Delete(message_rgn);
    if (OpLog) TAP_Hdd_Fclose(OpLog);
    TSRCommanderErase();
    TAP_Print("\nClosing %s", tap_name);
    TAP_Exit();
    return 0;
}

//----------------------------------------------------------------------------
dword TAP_EventHandler(word event, dword param1, dword param2)
{
    dword state, subState;
    int tick;

    switch (event) {
        case EVT_IDLE:
            tick = TAP_GetTick();
            //
            //message window
            //
            if (message_showing) {
                if (tick > message_showing) {
                    TAP_Osd_Delete(message_rgn);
                    message_rgn = 0;
                    message_showing = 0;
                }
            }

            TSRCommanderWork();
            break;
        case EVT_SVCSTATUS:
            break;

        case EVT_KEY:
            TAP_GetState(&state, &subState);
//            TAP_Print("\nEVT_KEY, event=%x, param1=%x, param2=%x", event, param1, param2);
//            TAP_Print("\nstate = %s, subState = %s", get_str(state_str, state), get_str(substate_str, subState));
            if (main_menu.visible) {
                param1 = menu_handler(&main_menu, param1);
            }
            else
            if (tfp_menu.visible) {
                param1 = tfp_menu_handler(&tfp_menu, param1);
            }
            else {
                if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                    param1 = key_handler(param1, param2);
                }
            }
            break;

        default:
            break;
    }

    return param1;
}


//----------------------------------------------------------------------------
void ShowMenu2(type_menu *menu)
{
    int i;
    int num_items;
    int y,h;

    if (!menu->visible) {
        TAP_ExitNormal();
        menu->rgn = TAP_Osd_Create(0, 0, 720, 576, 0, FALSE);
        if ((short)menu->rgn == -1) return;
        TAP_Win_SetDefaultColor(&menu->win);
        num_items = menu->item[0].id;
        h = (num_items + 1)* 20;
        h = (h < 80) ? 80 : h;
        h = (h > 280) ? 280 : h;
        y = 248 - (num_items * 10);
        y = (y < 50) ? 50 : y;
        y = (y > 250) ? 250 : y;
        //x,y,width,height
        TAP_Win_Create(&menu->win, (word)menu->rgn, 200, y, 300, h, FALSE, TRUE);
        TAP_Win_SetTitle(&menu->win, ptrfix(menu->fixme_titlep), 0, FNT_Size_1622);
//        TAP_Win_SetTitle(&menu->win, menu->title, 0, FNT_Size_1622);
        for (i=1; i<=menu->item[0].id; i++) {
            TAP_Win_AddItem(&menu->win, menu->item[i].str);
        }
        TAP_Win_SetSelection(&menu->win, 0);
        menu->visible = 1;
    }
}

int menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int ret;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
			switch (selection) {
				case 0:
				    remove_patch = 0;
                    build_tfp_menu(remove_patch);
                    ShowMenu2(&tfp_menu);
				    return 0;
				case 1:
				    remove_patch = 1;
                    build_tfp_menu(remove_patch);
                    ShowMenu2(&tfp_menu);
				    return 0;
				case 2:
				    return 0;
				case 3:
				    return 0;
				case 4:
				    return 0;
				case 5:
				    return 0;
				case 6:
				    return 0;
				case 7:
				    return 0;
				case 8:
				    return 0;
				case 9:
    				return 0;

				case 10:
                    close();
				    return 0;
			}
            break;
        case RKEY_Recall:
			DeleteMenu(menu);
            return param1;
        case RKEY_Exit:
			DeleteMenu(menu);
            return 0;
        default:
	        TAP_Win_Action(&menu->win, param1);
    		return 0;
    }
    return param1;
}

int tfp_menu_handler(type_menu *menu, dword param1)
{
    int selection;
    int ret;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
            if (selection < 0x20) {
                ret = read_tfp(tmenu[selection], &PatchModule, fw_gp, remove_patch, (debug & 1)?1:0);
                if (ret < 0) {
                    //try again using generic code
                    ret = read_tfp(tmenu[selection], &PatchModule, 0, remove_patch, (debug & 1)?1:0);
                }
                return 0;
            }
            break;
        case RKEY_Recall:
            DeleteMenu(menu);
            return param1;
        case RKEY_Exit:
            DeleteMenu(menu);
            return 0;
        default:
            TAP_Win_Action(&menu->win, param1);
            return 0;
    }
    return param1;
}


int build_tfp_menu(int remove)
{
    int ret;
    int i,j,n;
    TYPE_File Ip;

    ret = TAP_Hdd_ChangeDir("/");
    ret = TAP_Hdd_ChangeDir("ProgramFiles");
    if (ret == 0) {
        if (debug & 2) TAP_Print("\nERROR: cannot cd to /ProgramFiles");
        return -1;
    }


    n = TAP_Hdd_FindFirst(&Ip);
//printf("\nfound %d files", n);
    j = 0;
    for (i=0; i<n; i++) {
        if (Ip.attr != 0xff) {
        }
        if (Ip.attr == ATTR_FOLDER) {
            //don't enter subdirectories
        }
        else
        //ignore except *.tfp
        if (strlen(Ip.name) > 4) {
            if (strncmp(&Ip.name[strlen(Ip.name)-4], ".tfp", 4) == 0) {

//printf("\n%d - %s", i, Ip.name);
                if (remove) {
                    sprintf(tfp_menu.item[1+j].str, "Remove %s", Ip.name);
                }
                else {
                    sprintf(tfp_menu.item[1+j].str, "Load %s", Ip.name);
                }
                strcpy(tmenu[j], Ip.name);
                j++;

            }
        }
        TAP_Hdd_FindNext(&Ip);
    }
    tfp_menu.item[0].id = j;
    return 0;
}

int key_handler(dword param1, dword param2)
{
    int ret;

    switch(param1) {
        //only if tsrcommander is not running
        case RKEY_Recall:
            if (!is_tsr_commander_running()) {
                ShowMenu(&main_menu);
                return 0;
            }
            break;
//        case RKEY_Red:
//            red();
//            return 0;
//            break;
//        case RKEY_Green:
//            green();
//            return 0;
//            break;
//        case RKEY_Yellow:
//            yellow();
//            return 0;
//            break;
//        case RKEY_Blue:
//            blue();
//            return 0;
//            break;

//re test
//        case RKEY_Up:
//            printf("\nUp, %x", param2);
//            break;
//        case RKEY_Down:
//            printf("\nDown, %x", param2);
//            break;
//        case RKEY_Left:
//            printf("\nLeft, %x", param2);
//            break;
//        case RKEY_Right:
//            printf("\nRight, %x", param2);
//            break;

        default:
//            if (param2 && (param1 != param2)) printf("\nKEY %x %x", param1, param2);
            break;
    }
    return param1;
}

//-----------------------------------------------------------------------------
//
//print multiline message to OSD
//  500=vertical position(0-576)
//  RGB(3,5,10)=background colour(dark blue)
//  200=hold time(2 seconds)
//  0=blocking
//
void oprintf(const void *fmt, ...)
{
    va_list argp;
    int newline_count;
    int i,y;

    va_start(argp, fmt);
    vsprintf(strLog, fmt, argp);
//    TAP_Print("\n%s", strLog);

    newline_count = 0;
    for (i=0; i<strlen(strLog); i++) {
        if (strLog[i] == 0x0a) newline_count++;
    }
    y = 500 - (newline_count * 22);

    ShowMessage_y(strLog, y, RGB(3,5,10), 500, 0);
    va_end(argp);

    return;
}

void ffprintf(TYPE_File *Op, const void *fmt, ...)
{
    va_list argp;
    int len;

    va_start(argp, fmt);
    vsprintf(strLog, fmt, argp);
//    TAP_Print("\n%s", strLog);

    len = strlen(strLog);
    TAP_Hdd_Fwrite(strLog, 1, len, Op);

    va_end(argp);

    return;
}

//-----------------------------------------------------------------------------
char ini_filename[] = "patch_loader.ini";

int ini_read()
{
    int ret;
    int i;
    char key[0x20];
    char filename[0x64];

    ret = INIOpenFile(ini_filename);
//    printf("\nINIOpenFile(), ret=%x", ret);
    if (ret == 0) return -1;


    debug = INIGetHexDWord("debug", 0x00000000, 0x00000000, 0xffffffff);
//    printf("\nINIGetHexDWord(\"debug\"), debug=%x", debug);

    for (i=0; i<0x20; i++) {
        //patch_0=test.tfp
        sprintf(key, "patch_%d", i);
        ret = INIGetString(key, filename, "test.tfp", 0x64);
//        printf("\nINIGetString(key=%s, value=%s, default=%s, maxlen=%x), ret=%x", key, filename, "test.tfp", 0x64, ret);

        if (ret && strncmp(filename, "test.tfp", 8)) {
            //load patch
            ret = read_tfp(filename, &PatchModule, fw_gp, 0, (debug & 1)?1:0);
            if (ret < 0) {
                //try again using generic code
                ret = read_tfp(filename, &PatchModule, 0, 0, (debug & 1)?1:0);
            }
        }
    }

    INICloseFile();
//    printf("\n");
    return 0;
}

//-----------------------------------------------------------------------------
//
//testing
//
//int red() {
//    return 0;
//}
//
//int green() {
//    return 0;
//}
//
//int yellow() {
//    return 0;
//}
//
//int blue() {
//    return 0;
//}

//-----------------------------------------------------------------------------
#define fread  TAP_Hdd_Fread
#define fwrite TAP_Hdd_Fwrite
#define fclose TAP_Hdd_Fclose

int read_tfp(char *filename, type_PatchModule *PatchModule, dword gp, int remove, int dump)
{
    int ret;
    TYPE_File *Ip;
    int rlen;
    int i,j,k;
//    int dummy;

    ret = TAP_Hdd_ChangeDir("/");
    ret = TAP_Hdd_ChangeDir("ProgramFiles");
    if (ret == 0) {
        if (debug & 2) TAP_Print("\nERROR: cannot cd to /ProgramFiles");
        return -1;
    }


    Ip = TAP_Hdd_Fopen(filename);
    if (Ip == 0) {
        oprintf("Error: %s: Cannot open file: %s", filename);
        printf("\nError: %s: Cannot open file: %s\n", filename);
        return -1;
    }

    rlen = fread(&PatchModule->Magic, 1, sizeof(dword), Ip);
    rlen = fread(&PatchModule->FileVersion, 1, sizeof(dword), Ip);
    rlen = fread(&PatchModule->Author, 1, 0x50, Ip);
    rlen = fread(&PatchModule->Name, 1, 0x50, Ip);
    rlen = fread(&PatchModule->email, 1, 0x50, Ip);
    rlen = fread(&PatchModule->EnglishDescr, 1, 0x100, Ip);
    rlen = fread(&PatchModule->GermanDescr, 1, 0x100, Ip);
    rlen = fread(&PatchModule->Version, 1, sizeof(char), Ip);
//    rlen = fread(&dummy, 3, sizeof(char), Ip);                       //!!! NrOfPatches = big endian...
//    rlen = fread(&PatchModule->NrOfPatches, 1, sizeof(char), Ip);
//    PatchModule->NrOfPatches = LE32(PatchModule->NrOfPatches);
    rlen = fread(&PatchModule->NrOfPatches, 1, sizeof(dword), Ip);

    if (dump) {
        printf("[Generic]\n");
        printf("Author=%s\n", PatchModule->Author);
        printf("email=%s\n", PatchModule->email);
        printf("Name=%s\n", PatchModule->Name);
        printf("GermanDescr=%s\n", PatchModule->GermanDescr);
        printf("EnglishDesc=%s\n", PatchModule->EnglishDescr);
        printf("Version=%d\n", PatchModule->Version);
        printf("NrOfPatches=%x\n", PatchModule->NrOfPatches);
    }

    if (PatchModule->Magic != LE32(0x1a504654)) {
        oprintf("Error: %s: Invalid file format, %08x - not TFP", filename, PatchModule->Magic);
        printf("\nError: %s: Invalid file format, %08x - not TFP\n", filename, PatchModule->Magic);
        fclose(Ip);
        return -1;
    }
//    if (PatchModule->FileVersion != 2) {
//        printf("\nError: Invalid version - %d\n", PatchModule->FileVersion);
//        fclose(Ip);
//        return -1;
//    }

    for(k=0; k<PatchModule->NrOfPatches; k++) {
        rlen = fread(&PatchModule->PatchDetails.DestinationGP, 1, sizeof(dword), Ip);
        PatchModule->PatchDetails.DestinationGP = LE32(PatchModule->PatchDetails.DestinationGP);
        rlen = fread(&PatchModule->PatchDetails.NrOfSegments, 1, sizeof(dword), Ip);
        PatchModule->PatchDetails.NrOfSegments = LE32(PatchModule->PatchDetails.NrOfSegments);
        rlen = fread(&PatchModule->PatchDetails.Comment, 1, 0x50, Ip);

        if (dump) {
            printf("\n");
            printf("[Patch%02d]\n", k+1);
            printf("Comment=%s\n", PatchModule->PatchDetails.Comment);
            printf("DestinationGP=%08x\n", PatchModule->PatchDetails.DestinationGP);
            printf("NrOfSegments=%d\n", PatchModule->PatchDetails.NrOfSegments);
        }

        for(j=0; j<PatchModule->PatchDetails.NrOfSegments && j<MAX_SEGMENTS; j++) {
            rlen = fread(&PatchModule->PatchDetails.PatchSegment[j].Address, 1, sizeof(dword), Ip);
            PatchModule->PatchDetails.PatchSegment[j].Address = LE32(PatchModule->PatchDetails.PatchSegment[j].Address);
            rlen = fread(&PatchModule->PatchDetails.PatchSegment[j].CodeLength, 1, sizeof(dword), Ip);
            PatchModule->PatchDetails.PatchSegment[j].CodeLength = LE32(PatchModule->PatchDetails.PatchSegment[j].CodeLength);

            rlen = fread(&PatchModule->PatchDetails.PatchSegment[j].OriginalCode, PatchModule->PatchDetails.PatchSegment[j].CodeLength, 4, Ip);
            for (i=0; i<PatchModule->PatchDetails.PatchSegment[j].CodeLength; i++) PatchModule->PatchDetails.PatchSegment[j].OriginalCode[i] = LE32(PatchModule->PatchDetails.PatchSegment[j].OriginalCode[i]);

            rlen = fread(&PatchModule->PatchDetails.PatchSegment[j].OriginalMask, PatchModule->PatchDetails.PatchSegment[j].CodeLength, 4, Ip);
            for (i=0; i<PatchModule->PatchDetails.PatchSegment[j].CodeLength; i++) PatchModule->PatchDetails.PatchSegment[j].OriginalMask[i] = LE32(PatchModule->PatchDetails.PatchSegment[j].OriginalMask[i]);

            rlen = fread(&PatchModule->PatchDetails.PatchSegment[j].PatchedCode, PatchModule->PatchDetails.PatchSegment[j].CodeLength, 4, Ip);
            for (i=0; i<PatchModule->PatchDetails.PatchSegment[j].CodeLength; i++) PatchModule->PatchDetails.PatchSegment[j].PatchedCode[i] = LE32(PatchModule->PatchDetails.PatchSegment[j].PatchedCode[i]);

            rlen = fread(&PatchModule->PatchDetails.PatchSegment[j].PatchMask, PatchModule->PatchDetails.PatchSegment[j].CodeLength, 4, Ip);
            for (i=0; i<PatchModule->PatchDetails.PatchSegment[j].CodeLength; i++) PatchModule->PatchDetails.PatchSegment[j].PatchMask[i] = LE32(PatchModule->PatchDetails.PatchSegment[j].PatchMask[i]);

            if (PatchModule->PatchDetails.DestinationGP == gp) {
                if (dump) {
                    printf("\n");
                    printf("[Patch%02d_Segment%02d]\n", k+1, j+1);
                    printf("Address=%x\n", PatchModule->PatchDetails.PatchSegment[j].Address);
                    printf("CodeLength=%d\n", PatchModule->PatchDetails.PatchSegment[j].CodeLength);
                    printf("OriginalCode=");
                    for (i=0; i<PatchModule->PatchDetails.PatchSegment[j].CodeLength; i++) printf("%08x ", PatchModule->PatchDetails.PatchSegment[j].OriginalCode[i]);
                    printf("\n");
                    printf("OriginalMask=");
                    for (i=0; i<PatchModule->PatchDetails.PatchSegment[j].CodeLength; i++) printf("%08x ", PatchModule->PatchDetails.PatchSegment[j].OriginalMask[i]);
                    printf("\n");
                    printf("PatchedCode=");
                    for (i=0; i<PatchModule->PatchDetails.PatchSegment[j].CodeLength; i++) printf("%08x ", PatchModule->PatchDetails.PatchSegment[j].PatchedCode[i]);
                    printf("\n");
                    printf("PatchMask=  ");
                    for (i=0; i<PatchModule->PatchDetails.PatchSegment[j].CodeLength; i++) printf("%08x ", PatchModule->PatchDetails.PatchSegment[j].PatchMask[i]);
                    printf("\n");
                }

//                printf("\nUsing segment %02d, patch: %s",  j+1, PatchModule->PatchDetails.Comment);

                //if generic patch match against firmware
                if (PatchModule->PatchDetails.PatchSegment[j].Address == 0) {
                    ret = find_patch(&PatchModule->PatchDetails.PatchSegment[j]);
                    if (ret < 0) {
                        printf("\nfailed to match generic patch, %s", filename);
                        oprintf("failed to match generic patch, %s", filename);
                        fclose(Ip);
                        return -1;
                    }
                    else
                     if (ret == 1) {
                        printf("\nmatched generic patch, already patched, %s", filename);
                        oprintf("matched generic patch, already patched, %s", filename);
                        fclose(Ip);
                        return -1;
                    }
                }

                ret = check_patch(&PatchModule->PatchDetails.PatchSegment[j]);
                if (remove == 0) {
                    if(ret == 0) {
//                        printf("\nsegment %d", j);

                        //apply the patch
                        printf("\napplying patch, %s", filename);
                        oprintf("applying patch, %s", filename);
//                        memcpy((dword *)PatchModule->PatchDetails.PatchSegment[j].OriginalCodeCode, (dword *)PatchModule->PatchDetails.PatchSegment[j].Address, PatchModule->PatchDetails.PatchSegment[j].CodeLength *4);
                        memcpy((dword *)PatchModule->PatchDetails.PatchSegment[j].Address, (dword *)PatchModule->PatchDetails.PatchSegment[j].PatchedCode, PatchModule->PatchDetails.PatchSegment[j].CodeLength *4);
                    }
                    else
                    if(ret == 1) {
                        printf("\nalready patched, %s", filename);
                        oprintf("already patched, %s", filename);
                    }
                }
                else
                if (remove == 1) {
                    if (ret == 1) {
//                        printf("\nsegment %d", j);

                        //remove the patch
//may not have the original ...
                        if (gp) {
                            printf("\nremoving patch, %s", filename);
                            oprintf("removing patch, %s", filename);
                            memcpy((dword *)PatchModule->PatchDetails.PatchSegment[j].Address, (dword *)PatchModule->PatchDetails.PatchSegment[j].OriginalCode, PatchModule->PatchDetails.PatchSegment[j].CodeLength *4);
                        }
                        else {
                            printf("\ncannot remove generic patch - reboot!");
                            oprintf("cannot remove generic patch - reboot!");
                        }
                    }
                    else
                    if (ret == 0) {
                        printf("\nnot already patched, %s", filename);
                        oprintf("not already patched, %s", filename);
                    }
                }

            }
        }

        //quit if done
        if (PatchModule->PatchDetails.DestinationGP == gp) {
            fclose(Ip);
            return 0;
        }
    }

    fclose(Ip);
    return -1;
}

int find_patch(type_PatchSegment *PatchSegment)
{
    int ret;
    int i;
    dword j;
    int matched;


    if (PatchSegment->Address) {
        return PatchSegment->Address;
    }


    for (j=0x80000000; j<0x80260000; j+=4) {
        for (i=0; i<PatchSegment->CodeLength; i++) {

            matched = 1;
            if (*(dword *)(j + 4*i) != (PatchSegment->OriginalCode[i] & PatchSegment->OriginalMask[i])) {
                matched = 0;
                break;
            }
//else printf("\n%08x = %08x", j+4*i, *(dword *)(j + 4*i));
        }
        if (matched) {
//            printf("\nmatched original, ready to patch, addr=%08x", j);
            PatchSegment->Address = j;
            return 0;
        }

        for (i=0; i<PatchSegment->CodeLength; i++) {
            matched = 1;
            if (*(dword *)(j + 4*i) != (PatchSegment->PatchedCode[i] & PatchSegment->PatchMask[i])) {
                matched = 0;
                break;
            }
        }
        if (matched) {
//            printf("\nmatched patch, already patched, addr=%08x", j);
            PatchSegment->Address = j;
            return 1;
        }
    }

//    printf("\nfailed to match patch");
    return -1;
}



int check_patch(type_PatchSegment *PatchSegment)
{
    int ret;
    int i,j,k;
    int matched;


    if (is_firmware(PatchSegment->Address)) {
    }
    else {
        return -1;
    }


    //does the firmware match the original
    matched = 1;
    for (i=0; i<PatchSegment->CodeLength; i++) {
        if (*(dword *)(PatchSegment->Address + 4*i) != PatchSegment->OriginalCode[i] & PatchSegment->OriginalMask[i]) {
            matched = 0;
            break;
        }
    }
    if (matched) {
//        printf("\nmatch original, ready to patch");
    }
    else {
        //is the firmware already patched
        matched = 1;
        for (i=0; i<PatchSegment->CodeLength; i++) {
            if (*(dword *)(PatchSegment->Address + 4*i) != PatchSegment->PatchedCode[i] & PatchSegment->PatchMask[i]) {
                matched = 0;
                break;
            }
        }
        if (matched) {
//            printf("\nmatch patch, already patched");
            return 1;
        }
        else {
            printf("\noops, firmware in unknown state");
            oprintf("oops, firmware in unknown state");
dump32(PatchSegment->Address, 4*PatchSegment->CodeLength);
            return -1;
        }
    }

    return 0;
}



